#include <immintrin.h>
#include <omp.h>
#include <iostream>

using namespace std;


typedef unsigned long long uint64;

inline __attribute__((__may_alias__)) double bit_and_double(double* d, double* f) {
  long long t = (*(long long*)d) & (*(long long*)f);
  return *(double*)&t;
}

inline __attribute__((__may_alias__)) double bit_or_double(double* d, double* f) {
  long long t = (*(long long*)d) | (*(long long*)f);
  return *(double*)&t;
}

inline __attribute__((__may_alias__)) double bit_xor_double(double* d, double* f) {
  long long t = (*(long long*)d) ^ (*(long long*)f);
  return *(double*)&t;
}


inline __attribute__((__may_alias__)) double bit_andnot_double(double* d, double* f) {
  long long t = (*(long long*)d) &~ (*(long long*)f);
  return *(double*)&t;
}


double test_dp_mac(double x,double y,uint64 iterations){
    register double r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,rA,rB,rC,rD,rE,rF;

    //  Generate starting data.
    r0 = x;
    r1 = y;

    r8 = -0.0;

    r2 = bit_xor_double(&r0,&r8);
    r3 = bit_or_double(&r0,&r8);
    r4 = bit_andnot_double(&r8,&r0);
    
    r5 = r1 * 0.37796447300922722721;
    r6 = r1 * 0.24253562503633297352;
    r7 = r1 * 4.1231056256176605498;
    r8 = r0 + 0.37796447300922722721;
    r9 = r1 + 0.24253562503633297352;
    rA = r0 - 4.1231056256176605498;
    rB = r1 - 4.1231056256176605498;

    rC = 1.4142135623730950488;
    rD = 1.7320508075688772935;
    rE = 0.57735026918962576451;
    rF = 0.70710678118654752440;

    uint64 iMASK = 0x800fffffffffffffull;
    double  MASK = *(double*)&iMASK;
    double  vONE = 1.0;

    uint64 c = 0;
    while (c < iterations){
        size_t i = 0;
        while (i < 1000){

            r0 = r0*rC;
            r1 = r1+rD;
            r2 = r2*rE;
            r3 = r3+rF;
            r4 = r4*rC;
            r5 = r5+rD;
            r6 = r6*rE;
            r7 = r7+rF;
            r8 = r8*rC;
            r9 = r9+rD;
            rA = rA*rE;
            rB = rB+rF;

	    r0 = r0+rF;
            r1 = r1*rE;
            r2 = r2+rD;
            r3 = r3*rC;
            r4 = r4+rF;
            r5 = r5*rE;
            r6 = r6+rD;
            r7 = r7*rC;
            r8 = r8+rF;
            r9 = r9*rE;
            rA = rA+rD;
            rB = rB*rC;

            r0 = r0*rC;
            r1 = r1+rD;
            r2 = r2*rE;
            r3 = r3+rF;
            r4 = r4*rC;
            r5 = r5+rD;
            r6 = r6*rE;
            r7 = r7+rF;
            r8 = r8*rC;
            r9 = r9+rD;
            rA = rA*rE;
            rB = rB+rF;

	    r0 = r0+rF;
            r1 = r1*rE;
            r2 = r2+rD;
            r3 = r3*rC;
            r4 = r4+rF;
            r5 = r5*rE;
            r6 = r6+rD;
            r7 = r7*rC;
            r8 = r8+rF;
            r9 = r9*rE;
            rA = rA+rD;
            rB = rB*rC;

            i++;
        }

	//  Need to renormalize to prevent denormal/overflow.
        r0 = bit_and_double(&r0,&MASK);
        r1 = bit_and_double(&r1,&MASK);
        r2 = bit_and_double(&r2,&MASK);
        r3 = bit_and_double(&r3,&MASK);
        r4 = bit_and_double(&r4,&MASK);
        r5 = bit_and_double(&r5,&MASK);
        r6 = bit_and_double(&r6,&MASK);
        r7 = bit_and_double(&r7,&MASK);
        r8 = bit_and_double(&r8,&MASK);
        r9 = bit_and_double(&r9,&MASK);
        rA = bit_and_double(&rA,&MASK);
        rB = bit_and_double(&rB,&MASK);
        r0 = bit_or_double(&r0,&vONE);
        r1 = bit_or_double(&r1,&vONE);
        r2 = bit_or_double(&r2,&vONE);
        r3 = bit_or_double(&r3,&vONE);
        r4 = bit_or_double(&r4,&vONE);
        r5 = bit_or_double(&r5,&vONE);
        r6 = bit_or_double(&r6,&vONE);
        r7 = bit_or_double(&r7,&vONE);
        r8 = bit_or_double(&r8,&vONE);
        r9 = bit_or_double(&r9,&vONE);
        rA = bit_or_double(&rA,&vONE);
        rB = bit_or_double(&rB,&vONE);

        c++;
    }

    r0 = r0+r1;
    r2 = r2+r3;
    r4 = r4+r5;
    r6 = r6+r7;
    r8 = r8+r9;
    rA = rA+rB;

    r0 = r0+r2;
    r4 = r4+r6;
    r8 = r8+rA;

    r0 = r0+r4;
    r0 = r0+r8;
    

    return r0;
}

void test_dp_mac(int tds,uint64 iterations){

    double *sum = (double*)malloc(tds * sizeof(double));
    double start = omp_get_wtime();

#pragma omp parallel num_threads(tds)
    {
        double ret = test_dp_mac(1.1,2.1,iterations);
        sum[omp_get_thread_num()] = ret;
    }

    double secs = omp_get_wtime() - start;
    uint64 ops = 48 * 1000 * iterations * tds;
    cout << "Seconds     = " << secs << endl;
    cout << "FP Ops      = " << ops << endl;
    cout << "GFLOP/sec   = " << (ops / secs) / 1.0e9 << endl;

    double out = 0;
    int c = 0;
    while (c < tds){
        out += sum[c++];
    }

    cout << "sum = " << out << endl;
    cout << endl;

    free(sum);
}

int main(){
    //  (threads, iterations)
    test_dp_mac(4,1000000);

    //system("pause");
}
